@@ -1,5 +1,3 @@
-from plain.exceptions import ObjectDoesNotExist
-
 from . import (
     preflight,  # noqa
 )
@@ -65,6 +63,7 @@ from .registry import models_registry, register_model
 
 # Imports that would create circular imports if sorted
 from .base import DEFERRED, Model  # isort:skip
+from .options import Options  # isort:skip
 from .fields.related import (  # isort:skip
     ForeignKey,
     ManyToManyField,
@@ -78,7 +77,6 @@ from .fields.reverse_related import (  # isort:skip
 
 __all__ = aggregates_all + constraints_all + enums_all + fields_all + indexes_all
 __all__ += [
-    "ObjectDoesNotExist",
     "CASCADE",
     "DO_NOTHING",
     "PROTECT",
@@ -107,6 +105,7 @@ __all__ += [
     "JSONField",
     "Lookup",
     "Transform",
+    "Options",
     "Prefetch",
     "Q",
     "QuerySet",
