@@ -1,4 +1,4 @@
-def setup():
+def setup() -> None:
     # This package isn't an installed app,
     # so we need to trigger our own import and cli registration.
     from .cli import cli  # noqa
