@@ -1,5 +1,14 @@
+from __future__ import annotations
+
+from collections.abc import Callable
+from typing import TYPE_CHECKING, Any, cast
+
 from .base import Operation
 
+if TYPE_CHECKING:
+    from plain.models.backends.base.schema import BaseDatabaseSchemaEditor
+    from plain.models.migrations.state import ProjectState
+
 
 class SeparateDatabaseAndState(Operation):
     """
@@ -11,23 +20,33 @@ class SeparateDatabaseAndState(Operation):
 
     serialization_expand_args = ["database_operations", "state_operations"]
 
-    def __init__(self, database_operations=None, state_operations=None):
+    def __init__(
+        self,
+        database_operations: list[Operation] | None = None,
+        state_operations: list[Operation] | None = None,
+    ) -> None:
         self.database_operations = database_operations or []
         self.state_operations = state_operations or []
 
-    def deconstruct(self):
-        kwargs = {}
+    def deconstruct(self) -> tuple[str, list[Any], dict[str, list[Operation]]]:
+        kwargs: dict[str, list[Operation]] = {}
         if self.database_operations:
             kwargs["database_operations"] = self.database_operations
         if self.state_operations:
             kwargs["state_operations"] = self.state_operations
         return (self.__class__.__qualname__, [], kwargs)
 
-    def state_forwards(self, package_label, state):
+    def state_forwards(self, package_label: str, state: ProjectState) -> None:
         for state_operation in self.state_operations:
             state_operation.state_forwards(package_label, state)
 
-    def database_forwards(self, package_label, schema_editor, from_state, to_state):
+    def database_forwards(
+        self,
+        package_label: str,
+        schema_editor: BaseDatabaseSchemaEditor,
+        from_state: ProjectState,
+        to_state: ProjectState,
+    ) -> None:
         # We calculate state separately in here since our state functions aren't useful
         for database_operation in self.database_operations:
             to_state = from_state.clone()
@@ -37,7 +56,7 @@ class SeparateDatabaseAndState(Operation):
             )
             from_state = to_state
 
-    def describe(self):
+    def describe(self) -> str:
         return "Custom state/database change combination"
 
 
@@ -49,42 +68,68 @@ class RunSQL(Operation):
     by this SQL change, in case it's custom column/table creation/deletion.
     """
 
-    def __init__(self, sql, *, state_operations=None, elidable=False):
+    def __init__(
+        self,
+        sql: str
+        | list[str | tuple[str, list[Any]]]
+        | tuple[str | tuple[str, list[Any]], ...],
+        *,
+        state_operations: list[Operation] | None = None,
+        elidable: bool = False,
+    ) -> None:
         self.sql = sql
         self.state_operations = state_operations or []
         self.elidable = elidable
 
-    def deconstruct(self):
-        kwargs = {
+    def deconstruct(self) -> tuple[str, list[Any], dict[str, Any]]:
+        kwargs: dict[str, Any] = {
             "sql": self.sql,
         }
         if self.state_operations:
             kwargs["state_operations"] = self.state_operations
         return (self.__class__.__qualname__, [], kwargs)
 
-    def state_forwards(self, package_label, state):
+    def state_forwards(self, package_label: str, state: ProjectState) -> None:
         for state_operation in self.state_operations:
             state_operation.state_forwards(package_label, state)
 
-    def database_forwards(self, package_label, schema_editor, from_state, to_state):
+    def database_forwards(
+        self,
+        package_label: str,
+        schema_editor: BaseDatabaseSchemaEditor,
+        from_state: ProjectState,
+        to_state: ProjectState,
+    ) -> None:
         self._run_sql(schema_editor, self.sql)
 
-    def describe(self):
+    def describe(self) -> str:
         return "Raw SQL operation"
 
-    def _run_sql(self, schema_editor, sqls):
+    def _run_sql(
+        self,
+        schema_editor: BaseDatabaseSchemaEditor,
+        sqls: str
+        | list[str | tuple[str, list[Any]]]
+        | tuple[str | tuple[str, list[Any]], ...],
+    ) -> None:
         if isinstance(sqls, list | tuple):
-            for sql in sqls:
-                params = None
-                if isinstance(sql, list | tuple):
-                    elements = len(sql)
+            for sql_item in sqls:
+                params: list[Any] | None = None
+                sql: str
+                if isinstance(sql_item, list | tuple):
+                    elements = len(sql_item)
                     if elements == 2:
-                        sql, params = sql
+                        sql, params = sql_item  # type: ignore[misc]
                     else:
                         raise ValueError("Expected a 2-tuple but got %d" % elements)  # noqa: UP031
+                else:
+                    sql = sql_item
                 schema_editor.execute(sql, params=params)
         else:
-            statements = schema_editor.connection.ops.prepare_sql_script(sqls)
+            # sqls is a str in this branch
+            statements = schema_editor.connection.ops.prepare_sql_script(
+                cast(str, sqls)
+            )
             for statement in statements:
                 schema_editor.execute(statement, params=None)
 
@@ -96,7 +141,13 @@ class RunPython(Operation):
 
     reduces_to_sql = False
 
-    def __init__(self, code, *, atomic=None, elidable=False):
+    def __init__(
+        self,
+        code: Callable[..., Any],
+        *,
+        atomic: bool | None = None,
+        elidable: bool = False,
+    ) -> None:
         self.atomic = atomic
         # Forwards code
         if not callable(code):
@@ -104,20 +155,26 @@ class RunPython(Operation):
         self.code = code
         self.elidable = elidable
 
-    def deconstruct(self):
-        kwargs = {
+    def deconstruct(self) -> tuple[str, list[Any], dict[str, Any]]:
+        kwargs: dict[str, Any] = {
             "code": self.code,
         }
         if self.atomic is not None:
             kwargs["atomic"] = self.atomic
         return (self.__class__.__qualname__, [], kwargs)
 
-    def state_forwards(self, package_label, state):
+    def state_forwards(self, package_label: str, state: Any) -> None:
         # RunPython objects have no state effect. To add some, combine this
         # with SeparateDatabaseAndState.
         pass
 
-    def database_forwards(self, package_label, schema_editor, from_state, to_state):
+    def database_forwards(
+        self,
+        package_label: str,
+        schema_editor: BaseDatabaseSchemaEditor,
+        from_state: ProjectState,
+        to_state: ProjectState,
+    ) -> None:
         # RunPython has access to all models. Ensure that all models are
         # reloaded in case any are delayed.
         from_state.clear_delayed_models_cache()
@@ -127,5 +184,5 @@ class RunPython(Operation):
         # use direct imports, so we go with a documentation approach instead.
         self.code(from_state.models_registry, schema_editor)
 
-    def describe(self):
+    def describe(self) -> str:
         return "Raw Python operation"
