@@ -1,11 +1,19 @@
+from __future__ import annotations
+
+from typing import TYPE_CHECKING, Any
+
 from plain import models
 from plain.models.db import DatabaseError
+from plain.models.meta import Meta
 from plain.models.registry import ModelsRegistry
 from plain.utils.functional import classproperty
 from plain.utils.timezone import now
 
 from .exceptions import MigrationSchemaMissing
 
+if TYPE_CHECKING:
+    from plain.models.backends.base.base import BaseDatabaseWrapper
+
 
 class MigrationRecorder:
     """
@@ -20,10 +28,10 @@ class MigrationRecorder:
     a row in the table always means a migration is applied.
     """
 
-    _migration_class = None
+    _migration_class: type[models.Model] | None = None
 
     @classproperty
-    def Migration(cls):
+    def Migration(cls) -> type[models.Model]:  # type: ignore[misc]
         """
         Lazy load to avoid PackageRegistryNotReady if installed packages import
         MigrationRecorder.
@@ -37,31 +45,34 @@ class MigrationRecorder:
                 name = models.CharField(max_length=255)
                 applied = models.DateTimeField(default=now)
 
-                class Meta:
-                    models_registry = _models_registry
-                    package_label = "migrations"
-                    db_table = "plainmigrations"
+                # Use isolated models registry for migrations
+                _model_meta = Meta(models_registry=_models_registry)
 
-                def __str__(self):
+                model_options = models.Options(
+                    package_label="migrations",
+                    db_table="plainmigrations",
+                )
+
+                def __str__(self) -> str:
                     return f"Migration {self.name} for {self.app}"
 
             cls._migration_class = Migration
         return cls._migration_class
 
-    def __init__(self, connection):
+    def __init__(self, connection: BaseDatabaseWrapper) -> None:
         self.connection = connection
 
     @property
-    def migration_qs(self):
+    def migration_qs(self) -> Any:
         return self.Migration.query.all()
 
-    def has_table(self):
+    def has_table(self) -> bool:
         """Return True if the plainmigrations table exists."""
         with self.connection.cursor() as cursor:
             tables = self.connection.introspection.table_names(cursor)
-        return self.Migration._meta.db_table in tables
+        return self.Migration.model_options.db_table in tables
 
-    def ensure_schema(self):
+    def ensure_schema(self) -> None:
         """Ensure the table exists and has the correct schema."""
         # If the table's there, that's fine - we've never changed its schema
         # in the codebase.
@@ -76,7 +87,7 @@ class MigrationRecorder:
                 f"Unable to create the plainmigrations table ({exc})"
             )
 
-    def applied_migrations(self):
+    def applied_migrations(self) -> dict[tuple[str, str], Any]:
         """
         Return a dict mapping (package_name, migration_name) to Migration instances
         for all applied migrations.
@@ -91,16 +102,16 @@ class MigrationRecorder:
             # are applied.
             return {}
 
-    def record_applied(self, app, name):
+    def record_applied(self, app: str, name: str) -> None:
         """Record that a migration was applied."""
         self.ensure_schema()
         self.migration_qs.create(app=app, name=name)
 
-    def record_unapplied(self, app, name):
+    def record_unapplied(self, app: str, name: str) -> None:
         """Record that a migration was unapplied."""
         self.ensure_schema()
         self.migration_qs.filter(app=app, name=name).delete()
 
-    def flush(self):
+    def flush(self) -> None:
         """Delete all migration records. Useful for testing migrations."""
         self.migration_qs.all().delete()
