@@ -2,14 +2,21 @@
 Code to manage the creation and SQL rendering of 'where' constraints.
 """
 
+from __future__ import annotations
+
 import operator
 from functools import cached_property, reduce
+from typing import TYPE_CHECKING, Any
 
-from plain.exceptions import EmptyResultSet, FullResultSet
+from plain.models.exceptions import EmptyResultSet, FullResultSet
 from plain.models.expressions import Case, When
 from plain.models.lookups import Exact
 from plain.utils import tree
 
+if TYPE_CHECKING:
+    from plain.models.backends.base.base import BaseDatabaseWrapper
+    from plain.models.sql.compiler import SQLCompiler
+
 # Connection types
 AND = "AND"
 OR = "OR"
@@ -35,7 +42,9 @@ class WhereNode(tree.Node):
     resolved = False
     conditional = True
 
-    def split_having_qualify(self, negated=False, must_group_by=False):
+    def split_having_qualify(
+        self, negated: bool = False, must_group_by: bool = False
+    ) -> tuple[WhereNode | None, WhereNode | None, WhereNode | None]:
         """
         Return three possibly None nodes: one for those parts of self that
         should be included in the WHERE clause, one for those parts of self
@@ -111,7 +120,9 @@ class WhereNode(tree.Node):
         )
         return where_node, having_node, qualify_node
 
-    def as_sql(self, compiler, connection):
+    def as_sql(
+        self, compiler: SQLCompiler, connection: BaseDatabaseWrapper
+    ) -> tuple[str, list[Any]]:
         """
         Return the SQL version of the where clause and the value to be
         substituted in. Return '', [] if this node matches everything,
@@ -181,20 +192,20 @@ class WhereNode(tree.Node):
             sql_string = f"({sql_string})"
         return sql_string, result_params
 
-    def get_group_by_cols(self):
+    def get_group_by_cols(self) -> list[Any]:
         cols = []
         for child in self.children:
             cols.extend(child.get_group_by_cols())
         return cols
 
-    def get_source_expressions(self):
+    def get_source_expressions(self) -> list[Any]:
         return self.children[:]
 
-    def set_source_expressions(self, children):
+    def set_source_expressions(self, children: list[Any]) -> None:
         assert len(children) == len(self.children)
         self.children = children
 
-    def relabel_aliases(self, change_map):
+    def relabel_aliases(self, change_map: dict[str, str]) -> None:
         """
         Relabel the alias values of any children. 'change_map' is a dictionary
         mapping old (current) alias values to the new values.
@@ -206,7 +217,7 @@ class WhereNode(tree.Node):
             elif hasattr(child, "relabeled_clone"):
                 self.children[pos] = child.relabeled_clone(change_map)
 
-    def clone(self):
+    def clone(self) -> WhereNode:
         clone = self.create(connector=self.connector, negated=self.negated)
         for child in self.children:
             if hasattr(child, "clone"):
@@ -214,12 +225,12 @@ class WhereNode(tree.Node):
             clone.children.append(child)
         return clone
 
-    def relabeled_clone(self, change_map):
+    def relabeled_clone(self, change_map: dict[str, str]) -> WhereNode:
         clone = self.clone()
         clone.relabel_aliases(change_map)
         return clone
 
-    def replace_expressions(self, replacements):
+    def replace_expressions(self, replacements: dict[Any, Any]) -> WhereNode:
         if replacement := replacements.get(self):
             return replacement
         clone = self.create(connector=self.connector, negated=self.negated)
@@ -227,44 +238,44 @@ class WhereNode(tree.Node):
             clone.children.append(child.replace_expressions(replacements))
         return clone
 
-    def get_refs(self):
+    def get_refs(self) -> set[Any]:
         refs = set()
         for child in self.children:
             refs |= child.get_refs()
         return refs
 
     @classmethod
-    def _contains_aggregate(cls, obj):
+    def _contains_aggregate(cls, obj: Any) -> bool:
         if isinstance(obj, tree.Node):
             return any(cls._contains_aggregate(c) for c in obj.children)
         return obj.contains_aggregate
 
     @cached_property
-    def contains_aggregate(self):
+    def contains_aggregate(self) -> bool:
         return self._contains_aggregate(self)
 
     @classmethod
-    def _contains_over_clause(cls, obj):
+    def _contains_over_clause(cls, obj: Any) -> bool:
         if isinstance(obj, tree.Node):
             return any(cls._contains_over_clause(c) for c in obj.children)
         return obj.contains_over_clause
 
     @cached_property
-    def contains_over_clause(self):
+    def contains_over_clause(self) -> bool:
         return self._contains_over_clause(self)
 
     @property
-    def is_summary(self):
+    def is_summary(self) -> bool:
         return any(child.is_summary for child in self.children)
 
     @staticmethod
-    def _resolve_leaf(expr, query, *args, **kwargs):
+    def _resolve_leaf(expr: Any, query: Any, *args: Any, **kwargs: Any) -> Any:
         if hasattr(expr, "resolve_expression"):
             expr = expr.resolve_expression(query, *args, **kwargs)
         return expr
 
     @classmethod
-    def _resolve_node(cls, node, query, *args, **kwargs):
+    def _resolve_node(cls, node: Any, query: Any, *args: Any, **kwargs: Any) -> None:
         if hasattr(node, "children"):
             for child in node.children:
                 cls._resolve_node(child, query, *args, **kwargs)
@@ -273,23 +284,25 @@ class WhereNode(tree.Node):
         if hasattr(node, "rhs"):
             node.rhs = cls._resolve_leaf(node.rhs, query, *args, **kwargs)
 
-    def resolve_expression(self, *args, **kwargs):
+    def resolve_expression(self, *args: Any, **kwargs: Any) -> WhereNode:
         clone = self.clone()
         clone._resolve_node(clone, *args, **kwargs)
         clone.resolved = True
         return clone
 
     @cached_property
-    def output_field(self):
+    def output_field(self) -> Any:
         from plain.models.fields import BooleanField
 
         return BooleanField()
 
     @property
-    def _output_field_or_none(self):
+    def _output_field_or_none(self) -> Any:
         return self.output_field
 
-    def select_format(self, compiler, sql, params):
+    def select_format(
+        self, compiler: SQLCompiler, sql: str, params: list[Any]
+    ) -> tuple[str, list[Any]]:
         # Wrap filters with a CASE WHEN expression if a database backend
         # (e.g. Oracle) doesn't support boolean expression in SELECT or GROUP
         # BY list.
@@ -297,13 +310,13 @@ class WhereNode(tree.Node):
             sql = f"CASE WHEN {sql} THEN 1 ELSE 0 END"
         return sql, params
 
-    def get_db_converters(self, connection):
+    def get_db_converters(self, connection: BaseDatabaseWrapper) -> list[Any]:
         return self.output_field.get_db_converters(connection)
 
-    def get_lookup(self, lookup):
+    def get_lookup(self, lookup: str) -> Any:
         return self.output_field.get_lookup(lookup)
 
-    def leaves(self):
+    def leaves(self) -> Any:
         for child in self.children:
             if isinstance(child, WhereNode):
                 yield from child.leaves()
@@ -317,7 +330,11 @@ class NothingNode:
     contains_aggregate = False
     contains_over_clause = False
 
-    def as_sql(self, compiler=None, connection=None):
+    def as_sql(
+        self,
+        compiler: SQLCompiler | None = None,
+        connection: BaseDatabaseWrapper | None = None,
+    ) -> tuple[str, list[Any]]:
         raise EmptyResultSet
 
 
@@ -326,11 +343,15 @@ class ExtraWhere:
     contains_aggregate = False
     contains_over_clause = False
 
-    def __init__(self, sqls, params):
+    def __init__(self, sqls: list[str], params: list[Any] | None):
         self.sqls = sqls
         self.params = params
 
-    def as_sql(self, compiler=None, connection=None):
+    def as_sql(
+        self,
+        compiler: SQLCompiler | None = None,
+        connection: BaseDatabaseWrapper | None = None,
+    ) -> tuple[str, list[Any]]:
         sqls = [f"({sql})" for sql in self.sqls]
         return " AND ".join(sqls), list(self.params or ())
 
@@ -341,14 +362,18 @@ class SubqueryConstraint:
     contains_aggregate = False
     contains_over_clause = False
 
-    def __init__(self, alias, columns, targets, query_object):
+    def __init__(
+        self, alias: str, columns: list[str], targets: list[Any], query_object: Any
+    ):
         self.alias = alias
         self.columns = columns
         self.targets = targets
         query_object.clear_ordering(clear_default=True)
         self.query_object = query_object
 
-    def as_sql(self, compiler, connection):
+    def as_sql(
+        self, compiler: SQLCompiler, connection: BaseDatabaseWrapper
+    ) -> tuple[str, list[Any]]:
         query = self.query_object
         query.set_values(self.targets)
         query_compiler = query.get_compiler()
